# Copyright 2011 Canonical Ltd
#
from ConfigParser import ConfigParser
from argparse import ArgumentParser
import httplib2
from kanban import (
    LeankitKanban,
    Record,
    )
from launchpadlib.launchpad import Launchpad
import os.path
import re
import sys


TITLE_MARKER = "sync:"
DESCRIPTION_MARKER = "Synced with Launchpad"
NO_SYNC_MARKER = "(no-sync)"


def process_args(arglist=None):
    parser = ArgumentParser(
        description="Update a Kanban board from Launchpad data.")
    parser.add_argument(
        '-c', '--config', default='config.ini',
        help="LeankitKanban configuration file.")
    parser.add_argument(
        '-e', '--email', required=False, default=None,
        help="The Kanban user's registered email address")
    parser.add_argument(
        '-p', '--password', required=False, default=None,
        help="The user's Kanban password")
    parser.add_argument(
        '-b', '--board', required=False,
        help="The Kanban board (a title).")
    parser.add_argument(
        '-l', '--list', action='store_const', const=True, default=False,
        help="List available kanban boards.")
    parser.add_argument(
        '-d', '--debug', action='store_const', const=True, default=False,
        help="Turn on debugging for HTTP traffic")
    return parser.parse_args(arglist)


def parse_config(config):
    """Parse the kanban configuration from `config`.

    GLOBAL section is used for global settings for accessing the kanban
    board.  DEFAULT contains the default values which are included in
    each of the board sections automatically.

    Other sections should be named after the board title that they
    are configuring.
    """
    globals = {}
    for key, value in config.items('GLOBAL'):
        globals[key] = value

    boards = {}
    for section in config.sections():
        if section in ['GLOBAL']:
            continue
        boards[section] = {}
        for key in config.defaults():
            boards[section][key] = config.get(section, key)
        for key in config.options(section):
            boards[section][key] = config.get(section, key)
    return globals, boards


class LaunchpadUsersForBoard(object):
    """Find Launchpad user accounts for board users and keep them cached.

    Provides two dicts:
     * lp_to_kanban which uses Launchpad account names as keys and points
       to LeankitUser records, and
     * kanban_to_lp which uses LeankitUser records as keys pointing to
       actual Launchpad user records.
    """
    def __init__(self, lp, board):
        self.kanban_to_lp = {}
        self.lp_to_kanban = {}
        for user in board.users:
            lp_user = lp.people.getByEmail(email=user)
            assert lp_user is not None, (
                "There is no matching Launchpad user for kanban user %s." % (
                    user))
            self.lp_to_kanban[lp_user.name] = board.users[user]
            self.kanban_to_lp[board.users[user]] = lp_user


def get_branch_info(branches):
    """Returns a bug's status based on the given branches.

    Result will be one of "Approved", "Merged", "In Review", "In
    Progress", or None.  The meanings of the result values should be
    obvious with the addition that None signifies that no information
    could be gleaned from the branches.
    """
    info = Record(status=None, target=None, mp_url=None, mp_type=None)
    if len(branches) > 1:
        # We don't know what to do if the bug has more than one branch, so
        # refuse to guess.
        return info

    ORDERED_STATUSES = {'Rejected': 0,
                        'Work in progress': 0,
                        'Needs review': 1,
                        'Approved': 2,
                        'Merged': 2,
                        }
    REVIEW_STATUS = 'In Review'
    TRANSFORM_STATUS = {'Rejected': 'In Progress',
                        'Work in progress': 'In Progress',
                        'Needs review': REVIEW_STATUS,
                        'Approved': 'Approved',
                        'Merged': 'Merged',
                        }
    FINAL_STATUSES = ('Approved', 'Merged')

    def _get_mp_info():
        mps = []
        for bug_branch in branches:
            mp_info = Record(rank=None, status=None, mp=None)
            for mp in bug_branch.branch.landing_targets:
                status = mp.queue_status
                mp_info.rank = ORDERED_STATUSES.get(status, 1)
                mp_info.status = TRANSFORM_STATUS.get(status, REVIEW_STATUS)
                mp_info.mp = mp
                # If a merge proposal is in a final state, use it and stop
                # looking.
                if status in FINAL_STATUSES:
                    return mp_info
                mps.append(mp_info)
        if len(mps) == 0:
            return None
        mps.sort(key=lambda x:x.rank, reverse=True)
        return mps[0]

    def _get_mp_link(mp):
        regex = re.compile('(https://codereview.appspot.com/\d+)')
        match = regex.search(mp.description)
        if match:
            # An appspot Rietveld URL was found.  Use it.
            url = match.group(0)
            mp_type = 'Rietveld'
        else:
            url = mp.web_link
            mp_type = 'LP Merge Proposal'
        return (url, mp_type)

    mp_info = _get_mp_info()
    if mp_info is None:
        return info
    mp = mp_info.mp
    info.status = mp_info.status
    info.target = mp.target_branch.name
    info.mp_url, info.mp_type = _get_mp_link(mp)
    return info


def should_sync_card(card, conf={}):
    """Check to see if the card should be synced.

    `conf` is the kanban board configuration as a dict, and should contain
    an option `autosync` set to 'on' if all cards with external card IDs
    should be synced unless their description or title contain NO_SYNC_MARKER.
    With autosync off, if the card title starts with TITLE_MARKER or the card
    description contains DESCRIPTION_MARKER then return True.
    """
    if conf.get('autosync', None) == 'on':
        has_id = (card.external_card_id is not None and
                  card.external_card_id.strip() != '')
        return (has_id and NO_SYNC_MARKER not in card.title and
                NO_SYNC_MARKER not in card.description)
    else:
        return (card.title.startswith(TITLE_MARKER) or
                DESCRIPTION_MARKER in card.description)


NOOP_BUG_STATUSES = [
    "New", "Incomplete", "Confirmed", "Triaged",
]
IN_PROGRESS_BUG_STATUSES = [
    "In Progress", "Fix Committed",
    ]
DONE_BUG_STATUSES = [
    "Fix Released", "Opinion", "Invalid", "Won't Fix", "Expired",
]
ORDERED_BUG_STATUSES = (
    NOOP_BUG_STATUSES + IN_PROGRESS_BUG_STATUSES + DONE_BUG_STATUSES)


class CardStatus:
    """Card is consider to be in only one of these statuses."""
    CODING = 0
    REVIEW = 1
    LANDING = 2
    QA = 3
    DEPLOY = 4
    DOWNTIME_DEPLOY = 5
    DONE = 6


# Configuration variables for individual card statuses.
ConfigOptionByStatus = {
    CardStatus.CODING: 'coding_lanes',
    CardStatus.REVIEW: 'review_lanes',
    CardStatus.LANDING: 'landing_lanes',
    CardStatus.QA: 'qa_lanes',
    CardStatus.DEPLOY: 'deploy_lanes',
    CardStatus.DOWNTIME_DEPLOY: 'downtime_deploy_lanes',
    CardStatus.DONE: 'done_lanes',
    }


def get_card_status(bug_status, bug_tags, branch_info):
    """Return the status of the card as one of CardStatus values."""
    # status of None means card needs not to be moved.
    status = None
    if bug_status in IN_PROGRESS_BUG_STATUSES:
        status = CardStatus.CODING
        if branch_info.status == 'In Progress':
            status = CardStatus.CODING
        elif branch_info.status == 'In Review':
            status = CardStatus.REVIEW
        elif branch_info.status == 'Approved':
            status = CardStatus.LANDING
        elif branch_info.status == 'Merged':
            if bug_status == 'Fix Committed':
                if 'qa-needstesting' in bug_tags:
                    status = CardStatus.QA
                else:
                    if branch_info.target == 'db-devel':
                        status = CardStatus.DOWNTIME_DEPLOY
                    else:
                        status = CardStatus.DEPLOY
            else:
                status = CardStatus.LANDING
    elif bug_status in DONE_BUG_STATUSES:
        status = CardStatus.DONE
    return status


def get_bug_status(lp_bug, all_projects, all_users):
    lowest_status = None
    matches = False
    assignees = []
    tasks = lp_bug.bug_tasks_collection
    for task in tasks:
        if (task.target in all_projects or
            (task.assignee is not None and
             task.assignee.name in all_users)):
            matches = True
            if (lowest_status is None or
                (ORDERED_BUG_STATUSES.index(task.status) <
                 ORDERED_BUG_STATUSES.index(lowest_status))):
                lowest_status = task.status
            if task.assignee is not None:
                assignees.append(task.assignee)
    branch_info = get_branch_info(lp_bug.linked_branches)
    status = get_card_status(lowest_status, lp_bug.tags, branch_info)
    return matches, status, assignees, branch_info.mp_url, branch_info.mp_type


def find_card_target_lane(lane, path):
    if lane.path.endswith(path):
        return lane
    else:
        next_lanes = lane.getNextLanes()
        if next_lanes is None:
            return None

        for next_lane in next_lanes:
            target_lane = find_card_target_lane(next_lane, path)
            if target_lane is not None:
                return target_lane
        return None


def get_lp(config):
    """Get a lp instance for a logged in user or anonymous access.

    If 'credentials_file' is specified in the board configuration then an
    instance of Launchpad for the user is returned, otherwise anonymous access
    is used.
    """
    credentials_file = config.get('credentials_file')
    if credentials_file and os.path.exists(credentials_file):
        lp = Launchpad.login_with(
            'lp2kanban', config['lp_server'],
            credentials_file=credentials_file)
        print "   Logged in as", lp.me.name
    else:
        lp = Launchpad.login_anonymously('lp2kanban', config['lp_server'])
        print ("   Accessing Launchpad anonymously.  "
               "Private bugs will not be synced.")
    return lp


def sync_board(board, bconf):
    lp = get_lp(bconf)
    all_projects = []
    project_group = bconf.get('project_group')
    if project_group:
        launchpad_project_group = lp.project_groups[project_group]
        # These are sometimes very slow (and basically take a considerable
        # chunk of the runtime).  We should optimize this so it's done only
        # once per run instead of once per board being synced.
        all_projects = list(launchpad_project_group.projects_collection)
    project_names = bconf.get('projects')
    if project_names:
        for name in project_names.split(','):
            project = lp.projects[name.strip()]
            if project:
                all_projects.append(project)
    lp_users = LaunchpadUsersForBoard(lp, board)
    print "   Syncing cards:"
    for card in board.getCardsWithExternalIds():
        if should_sync_card(card, bconf):
            bug_id = int(card.external_card_id)
            try:
                lp_bug = lp.bugs[bug_id]
            except KeyError:
                print (
                    "   * #%d is not accessible "
                    "(does not exist or forbidden)." % (bug_id))
                continue
            synced, bug_status, assignees, mp_url, mp_type = get_bug_status(
                lp_bug, all_projects, lp_users.lp_to_kanban.keys())
            target_lane = None
            if bug_status is not None:
                target_lane_path = bconf[ConfigOptionByStatus[bug_status]]
                target_lane = find_card_target_lane(
                    card.lane, target_lane_path)
                try:
                    card.move(target_lane)
                except IOError as e:
                    print "   * #%s: %s (in %s)" % (
                        bug_id, card.title, card.lane.path)
                    print "     >>> Error moving card:"
                    print "     ", e
                    continue

            if synced:
                card.title = lp_bug.title
                for tag in lp_bug.tags:
                    card.addTag(tag)
                if len(assignees) > 0:
                    assignee = assignees[0]
                    kanban_user = lp_users.lp_to_kanban.get(
                        assignee.name, None)
                    if kanban_user:
                        card.assigned_user_id = kanban_user.id
                if mp_url and mp_type:
                    card.external_system_url = mp_url
                    card.external_system_name = mp_type
                print "   * #%s: %s (in %s)" % (
                    bug_id, card.title, card.lane.path)
            else:
                print "   * #%s: %s -- Not synced" % (
                    bug_id, card.title)


def save_board(board):
    # Split out so we can change the call to a single UpdateCards call.
    for card in board.cards:
        if card.is_dirty:
            if DESCRIPTION_MARKER not in card.description:
                card.description += '\n' + DESCRIPTION_MARKER
                card.description = card.description.strip()

            print "     Saving '%s' for #%s." % (
                "', '".join(card.dirty_attrs), card.external_card_id)
        card.save()


def main(argv):
    """Update the Yellow team's kanban board with data from Launchpad."""
    args = process_args()

    if args.debug:
        httplib2.debuglevel = 1

    config = ConfigParser()
    config.read(args.config)
    global_config, boards_config = parse_config(config)
    user = args.email or global_config['user']
    password = args.password or global_config['password']
    kanban = LeankitKanban(global_config['account'], user, password)

    if args.list:
        print "All active boards:"
        boards = kanban.getBoards()
        for board in boards:
            print " * %s (%d)" % (board.title, board.id)
        return

    if args.board:
        assert args.board in boards_config, (
            "Requested board '%s' is not present in the config file %s." % (
                args.board, args.config))
        board_names = [args.board]
    else:
        board_names = boards_config.keys()
    print "Requested boards:"
    for board_name in board_names:
        print " * %s" % board_name
        board = kanban.getBoard(title=board_name)
        if board is None:
            print "Could not retrieve board.  Skipping."
            continue
        sync_board(board, boards_config[board_name])
        save_board(board)


if __name__ == '__main__':
    main(sys.argv)
