__metaclass__ = type

from ConfigParser import (
    ConfigParser,
    NoSectionError,
    )
from StringIO import StringIO
from textwrap import dedent
import unittest

from lp2kanban.bugs2cards import (
    CardStatus,
    find_card_target_lane,
    get_branch_info,
    get_bug_status,
    get_card_status,
    parse_config,
    should_sync_card,
    )
from lp2kanban.kanban import Record

class ConfigReadingTest(unittest.TestCase):

    def test_parse_config_no_globals(self):
        # When 'GLOBAL' section is missing from the configuration,
        # an exception is raised.
        sample_conf = '''
            [Test]
            var = board value
        '''

        config = ConfigParser()
        config.readfp(StringIO(dedent(sample_conf)))

        # NoSectionError is raised because there is no obligatory GLOBAL
        # configuration section.
        self.assertRaises(NoSectionError, parse_config, config)

    def test_parse_config_globals(self):
        # Values in the 'GLOBAL' section are returned separately and
        # are independent from per-board configuration.
        sample_conf = '''
            [GLOBAL]
            var1 = value1
            var2 = value2
            [Board 1]
            var1 = board value 1
        '''

        config = ConfigParser()
        config.readfp(StringIO(dedent(sample_conf)))

        globals, boards = parse_config(config)

        self.assertEqual({'var1': 'value1',
                          'var2': 'value2',
                           }, globals)
        self.assertEqual({'Board 1': {'var1': 'board value 1'}}, boards)

    def test_parse_config_defaults(self):
        # Values from the DEFAULT section are always added to every
        # board configuration section as well.
        sample_conf = '''
            [GLOBAL]
            [DEFAULT]
            var = board value
            [Board 1]
        '''

        config = ConfigParser()
        config.readfp(StringIO(dedent(sample_conf)))

        globals, boards = parse_config(config)
        self.assertEqual({'Board 1': {'var': 'board value'}}, boards)

    def test_parse_config_defaults_overridden(self):
        # Each board can override settings from the defaults.
        sample_conf = '''
            [GLOBAL]
            [DEFAULT]
            var = board value
            [Board 1]
            var = my board value
        '''

        config = ConfigParser()
        config.readfp(StringIO(dedent(sample_conf)))

        globals, boards = parse_config(config)
        self.assertEqual({'Board 1': {'var': 'my board value'}}, boards)


class FauxBugBranch:
    def __init__(self, mps):
        self.branch = Record(landing_targets=mps)


class FauxMP:
    def __init__(self, queue_status, target_name, link=None, description=None):
        self.queue_status = queue_status
        self.target_branch = Record(name=target_name)
        self.web_link = link
        self.description = description or ''


class BranchInfoTests(unittest.TestCase):

    def test_no_guessing_on_multiple_branches(self):
        # If a bug has more than one branch associated with it we act as if it
        # has no branches because the code doesn't (yet) know how to interpret
        # more than one branch.
        result = get_branch_info([None, None])
        self.assertEqual(result, {'status': None, 'target': None,
                                  'mp_url': None, 'mp_type': None})

    def test_no_branches_mean_no_info(self):
        # If there are no branches associated with a bug, then no info is
        # returned.
        result = get_branch_info([])
        self.assertEqual(result, {'status': None, 'target': None,
                                  'mp_url': None, 'mp_type': None})

    def test_branches_returns_target(self):
        # The MP's target is pulled out of the bug branch and returned by
        # get_branch_info.
        branches = [FauxBugBranch([FauxMP('Approved', 'devel')])]
        result = get_branch_info(branches)
        self.assertEqual(result['target'], 'devel')

    def test_branches_returns_mp_link(self):
        # The MP's link is returned as the mp_url.
        branches = [FauxBugBranch([FauxMP('Approved', 'devel',
                                          'http://example.com/mp/link')])]
        result = get_branch_info(branches)
        self.assertEqual(result['mp_url'], 'http://example.com/mp/link')

    def test_branches_returns_rietveld_link(self):
        # If, however, a link to https://coderview.appspot.com/\d+ is found
        # in the MP description, then that URL is returned.
        branches = [FauxBugBranch([FauxMP(
            'Approved', 'devel', 'http://example.com/mp/link',
            'https://codereview.appspot.com/1234')])]
        result = get_branch_info(branches)
        self.assertEqual(result['mp_url'],
                         'https://codereview.appspot.com/1234')

    def test_branches_in_review(self):
        # If there is an MP then normally the bug's status is considered "In
        # Review".
        branches = [FauxBugBranch([FauxMP('Fake', 'devel')])]
        result = get_branch_info(branches)
        self.assertEqual(result['status'], 'In Review')

    def test_approved_branches_are_approved(self):
        branches = [FauxBugBranch([FauxMP('Approved', 'devel')])]
        result = get_branch_info(branches)
        self.assertEqual(result['status'], 'Approved')

    def test_merged_branches_are_merged(self):
        branches = [FauxBugBranch([FauxMP('Merged', 'devel')])]
        result = get_branch_info(branches)
        self.assertEqual(result['status'], 'Merged')

    def test_in_progress_branches_are_in_progress(self):
        branches = [FauxBugBranch([FauxMP('Work in progress', 'devel')])]
        result = get_branch_info(branches)
        self.assertEqual(result['status'], 'In Progress')

    def test_rejected_branches_are_in_progress(self):
        branches = [FauxBugBranch([FauxMP('Rejected', 'devel')])]
        result = get_branch_info(branches)
        self.assertEqual(result['status'], 'In Progress')

    def test_highest_ranked_returned(self):
        # If there are multiple merge proposals, the one with the highest rank
        # wins.
        branches = [FauxBugBranch([FauxMP('In Review', 'devel'),
                                   FauxMP('Approved', 'approved-branch'),
                                   FauxMP('Rejected', 'bad-branch'),
                                   ])]
        result = get_branch_info(branches)
        self.assertEqual(result['status'], 'Approved')

    def test_first_seen_final_state_returned(self):
        # If there are multiple MPs in a final state, the first one seen wins.
        branches = [FauxBugBranch([FauxMP('In Review', 'devel'),
                                   FauxMP('Merged', 'merged-branch'),
                                   FauxMP('Approved', 'approved-branch'),
                                   ])]
        result = get_branch_info(branches)
        self.assertEqual(result['status'], 'Merged')


class FauxBug:
    def __init__(self, tags=[]):
        self.bug_tasks_collection = []
        self.linked_branches = []
        self.tags = tags

    def addTask(self, target, status='New', assignee=None):
        self.bug_tasks_collection.append(
            Record(target=target, status=status, assignee=assignee))


class BugStatusTest(unittest.TestCase):

    def test_get_bug_status_unknown(self):
        # For a bug on an unrelated project and not assigned to any
        # of the relevant users, it is not considered to match at all.
        bug = FauxBug()
        bug.addTask('project')
        matches, status, assignees, mp_url, mp_type = (
            get_bug_status(bug, ['launchpad'], []))
        self.assertFalse(matches)
        self.assertEqual(None, status)
        self.assertEqual([], assignees)

    def test_get_bug_status_assignee(self):
        # For a bug on an unrelated project and assigned to a user
        # we care about, it is returned as matching with proper status.
        bug = FauxBug()
        user = Record(name='user')
        bug.addTask('project', status='In Progress', assignee=user)
        matches, status, assignees, mp_url, mp_type = (
            get_bug_status(bug, ['launchpad'], ['user']))
        self.assertTrue(matches)
        self.assertEqual(CardStatus.CODING, status)
        self.assertEqual([user], assignees)

    def test_get_bug_status_project(self):
        # For a bug on an project we care about, it is returned
        # as matching and with proper status.
        bug = FauxBug()
        nonuser = Record(name='nonuser')
        bug.addTask('launchpad', status='In Progress', assignee=nonuser)
        matches, status, assignees, mp_url, mp_type = (
            get_bug_status(bug, ['launchpad'], []))
        self.assertTrue(matches)
        self.assertEqual(CardStatus.CODING, status)
        self.assertEqual([nonuser], assignees)

    def test_get_bug_status_multi_targets(self):
        # For a bug with multiple tasks, only some of which are on
        # projects we care about, only assignees from those tasks are
        # returned.
        bug = FauxBug()
        nonuser = Record(name='nonuser')
        user = Record(name='user')
        bug.addTask('project', status='In Progress', assignee=nonuser)
        bug.addTask('launchpad', status='Fix Released', assignee=user)
        matches, status, assignees, mp_url, mp_type = (
            get_bug_status(bug, ['launchpad'], []))
        self.assertTrue(matches)
        self.assertEqual(CardStatus.DONE, status)
        self.assertEqual([user], assignees)

    def test_get_bug_status_multi_statuses(self):
        # For a bug with multiple tasks, all of which are on
        # projects we care about, the bug status is considered to be the
        # "lowest" of all bug task statuses (i.e. 'In Progress' is lower
        # than 'Fix Committed').
        bug = FauxBug()
        bug.addTask('lazr', status='Fix Released')
        bug.addTask('launchpad', status='In Progress')
        matches, status, assignees, mp_url, mp_type = get_bug_status(
            bug, ['launchpad', 'lazr'], [])
        self.assertTrue(matches)
        # 'Fix Released' on 'lazr' would put the bug status in DONE state,
        # but since there is an 'In Progress' task as well, it is instead
        # in CODING.
        self.assertEqual(CardStatus.CODING, status)


class CardStatusTest(unittest.TestCase):

    def test_should_sync_card_no(self):
        # Cards with sync markers in neither title nor description are
        # not synced.
        card = Record(title=u'no sync', description=u'really no sync')
        self.assertFalse(should_sync_card(card))

    def test_should_sync_card_title_yes(self):
        # Sync marker for the title is a string 'sync:' at the beginning
        # of the title string.
        card = Record(title=u'sync:', description=u'description')
        self.assertTrue(should_sync_card(card))

    def test_should_sync_card_title_no(self):
        # Sync marker for the title is a string 'sync:' at the beginning
        # of the title string, anywhere else and it's ignored.
        card = Record(title=u'no sync:', description=u'description')
        self.assertFalse(should_sync_card(card))

    def test_should_sync_card_description_yes(self):
        # Sync marker for the description is a string 'sync:' at the beginning
        # of the title string.
        card = Record(title=u'no sync', description=u'Synced with Launchpad')
        self.assertTrue(should_sync_card(card))

    def test_should_sync_card_autosync_no(self):
        # When autosync is 'on', cards with no external card IDs
        # are not synced.
        card = Record(title=u'no sync', description=u'',
                      external_card_id=None)
        self.assertFalse(should_sync_card(card))

    def test_should_sync_card_autosync_yes(self):
        # When autosync is 'on', cards with external card IDs are synced.
        card = Record(title=u'no sync', description=u'',
                      external_card_id='11')
        self.assertTrue(should_sync_card(card, {'autosync': 'on'}))

    def test_should_sync_card_autosync_nosync(self):
        # When autosync is 'on', cards with external card IDs are not
        # synced when they contain the no-sync marker.
        card = Record(title=u'(no-sync)', description=u'(no-sync)',
                      external_card_id='11')
        self.assertFalse(should_sync_card(card, {'autosync': 'on'}))

    def test_get_card_status_noop(self):
        # For a bug in 'New', 'Triaged', 'Confirmed' or 'Incomplete' statuses,
        # the status is unknown.
        self.assertEqual(None, get_card_status('New', [], None))
        self.assertEqual(None, get_card_status('Confirmed', [], None))
        self.assertEqual(None, get_card_status('Incomplete', [], None))
        self.assertEqual(None, get_card_status('Triaged', [], None))

    def test_get_card_status_coding_no_branch(self):
        # For a bug in 'In Progress' or 'Fix Committed' status, it is
        # considered to be in the coding state if there is no branch.
        branch_info = Record(status=None, target=None)
        self.assertEqual(
            CardStatus.CODING,
            get_card_status('In Progress', [], branch_info))
        self.assertEqual(
            CardStatus.CODING,
            get_card_status('Fix Committed', [], branch_info))

    def test_get_card_status_coding_branch(self):
        # For a bug in 'In Progress' or 'Fix Committed' status, it is
        # considered to be in the coding state if there is a branch
        # and it is 'In Progress'.
        branch_info = Record(status='In Progress', target=None)
        self.assertEqual(
            CardStatus.CODING,
            get_card_status('In Progress', [], branch_info))
        self.assertEqual(
            CardStatus.CODING,
            get_card_status('Fix Committed', [], branch_info))

    def test_get_card_status_review(self):
        # For a bug in 'In Progress' or 'Fix Committed' status, it is
        # considered to be in the review phase if there is a branch
        # proposed for review.
        branch_info = Record(status='In Review', target=None)
        self.assertEqual(
            CardStatus.REVIEW,
            get_card_status('In Progress', [], branch_info))
        self.assertEqual(
            CardStatus.REVIEW,
            get_card_status('Fix Committed', [], branch_info))

    def test_get_card_status_landing(self):
        # For a bug in 'In Progress' or 'Fix Committed' status, it is
        # considered to be in the landing phase if there is a branch
        # approved for landing.
        branch_info = Record(status='Approved', target=None)
        self.assertEqual(
            CardStatus.LANDING,
            get_card_status('In Progress', [], branch_info))
        self.assertEqual(
            CardStatus.LANDING,
            get_card_status('Fix Committed', [], branch_info))

    def test_get_card_status_qa(self):
        # For a bug in the 'Fix Committed' status, it is considered
        # to be in the QA phase if there is a branch that has been merged
        # and the bug has a 'qa-needstesting' tag.
        branch_info = Record(status='Merged', target=None)
        self.assertEqual(
            CardStatus.QA,
            get_card_status('Fix Committed', ['qa-needstesting'], branch_info))

        # If the bug is still 'In Progress', it is still in the landing phase.
        self.assertEqual(
            CardStatus.LANDING,
            get_card_status('In Progress', ['qa-needstesting'], branch_info))


    def test_get_card_status_deploy(self):
        # For a bug in the 'Fix Committed' status, it is
        # considered to be in the deploy phase if there is a branch
        # that has been merged and the bug has no 'qa-needstesting' tag,
        # and the branch target is 'devel'
        branch_info = Record(status='Merged', target='devel')
        self.assertEqual(
            CardStatus.DEPLOY,
            get_card_status('Fix Committed', [], branch_info))

        # If the bug is still 'In Progress', it is still in the landing phase.
        self.assertEqual(
            CardStatus.LANDING,
            get_card_status('In Progress', [], branch_info))

    def test_get_card_status_downtime_deploy(self):
        # For a bug in the 'Fix Committed' status, it is
        # considered to be in the downtime-deploy phase if there is a branch
        # that has been merged and the bug has no 'qa-needstesting' tag,
        # and the branch target is 'db-devel'
        branch_info = Record(status='Merged', target='db-devel')
        self.assertEqual(
            CardStatus.DOWNTIME_DEPLOY,
            get_card_status('Fix Committed', [], branch_info))

        # If the bug is still 'In Progress', it is still in the landing phase.
        self.assertEqual(
            CardStatus.LANDING,
            get_card_status('In Progress', [], branch_info))

    def test_get_card_status_done(self):
        # For a bug in 'Fix Released', 'Opinion', 'Won't fix', 'Invalid' or
        # or 'Expired' statuses, it is considered completely done.
        self.assertEqual(
            CardStatus.DONE, get_card_status('Fix Released', [], None))
        self.assertEqual(
            CardStatus.DONE, get_card_status('Opinion', [], None))
        self.assertEqual(
            CardStatus.DONE, get_card_status("Won't Fix", [], None))
        self.assertEqual(
            CardStatus.DONE, get_card_status('Invalid', [], None))
        self.assertEqual(
            CardStatus.DONE, get_card_status('Expired', [], None))


class FauxLane:
    def __init__(self, path):
        self.path = path
        self.next_lanes = None

    def getNextLanes(self):
        return self.next_lanes

    def addNextLane(self, lane):
        if self.next_lanes is None:
            self.next_lanes = [lane]
        else:
            self.next_lanes.append(lane)

class FindLaneTest(unittest.TestCase):

    def test_find_card_target_lane_no_match(self):
        # Lane path needs to end with the string we are looking for.
        coding = FauxLane('Dev::Coding')
        review = FauxLane('Dev::Review')
        coding.addNextLane(review)

        self.assertEqual(None, find_card_target_lane(coding, 'Landing'))

    def test_find_card_target_lane_exists(self):
        coding = FauxLane('Dev::Coding')
        review = FauxLane('Dev::Review')
        coding.addNextLane(review)
        landing = FauxLane('Dev::Landing')
        review.addNextLane(landing)

        self.assertEqual(landing, find_card_target_lane(coding, '::Landing'))

    def test_find_card_target_lane_branching(self):
        # All the potential next lanes are searched through recursively
        # for the matching next lane.
        qa = FauxLane('QA::Done')
        land = FauxLane('Land::Quick')
        slow_land = FauxLane('Land::Slow')
        qa.addNextLane(land)
        qa.addNextLane(slow_land)
        rollout = FauxLane('Rollout')
        slow_land.addNextLane(rollout)

        self.assertEqual(rollout, find_card_target_lane(qa, 'Rollout'))
